import unittest
import event

class TestUpgrader(unittest.TestCase):
    def test_multiple_versions(self):
        @event.upgrade('0.1', '0.2')
        def set_name(event):
            event['name'] = 'first upgrade'
        @event.upgrade('0.2', '0.3')
        def reset_name(event):
            event['name'] = 'second upgrade'
        upgrader = event.Upgrader([set_name, reset_name])
        e = {'test': 'one', 'version': '0.1'}
        u = upgrader.upgrade(e)
        self.assertEqual(u['name'], 'second upgrade')
        self.assertEqual(u['version'], '0.3')

    def test_exception_if_not_upgraded(self):
        @event.upgrade('0.1')
        def set_name(event):
            event['name'] = 'first upgrade'
        upgrader = event.Upgrader([set_name])
        e = {'test': 'one', 'version': '0.1'}
        with self.assertRaises(KeyError):
            u = upgrader.upgrade(e)

    def test_no_version(self):
        @event.upgrade('0.1', '0.2')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'
        upgrader = event.Upgrader([set_name])
        e = {'test': 'one'}
        u = upgrader.upgrade(e)
        self.assertEqual(u['test'], 'one')
        self.assertFalse('version' in u)
        self.assertFalse('name' in u)

    def test_target_version_failure(self):
        @event.upgrade('0.1', '0.2')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'
        upgrader = event.Upgrader([set_name])
        e = {'test': 'one', 'version': '0.1'}
        with self.assertRaises(KeyError):
            u = upgrader.upgrade(e, target_version='0.3')

    def test_target_version_break(self):
        @event.upgrade('0.1', '0.2')
        def set_name(event):
            event['name'] = 'first upgrade'
        @event.upgrade('0.2', '0.3')
        def reset_name(event):
            event['name'] = 'second upgrade'
        upgrader = event.Upgrader([set_name, reset_name])
        e = {'test': 'one', 'version': '0.1'}
        u = upgrader.upgrade(e, target_version='0.2')
        self.assertEqual(u['name'], 'first upgrade')
        self.assertEqual(u['version'], '0.2')

    def test_ingest_factory(self):
        @event.upgrade('0.1', '0.2')
        def set_name(event: dict):
            event['name'] = 'first upgrade'
        @event.upgrade('0.2', '0.3')
        def reset_name(event: dict):
            event['name'] = 'second upgrade'
        def factory(event: dict):
            return 'factory for {test} {name}'.format(**event)
        upgrader = event.Upgrader([set_name, reset_name], factory=factory)
        e = {'test': 'one', 'version': '0.1'}
        u = upgrader.ingest(e, target_version='0.3')
        self.assertEqual(u, 'factory for one second upgrade')


import unittest
import event

class TestEventWrapper(unittest.TestCase):
    def test_default_version(self):
        e = event.Event({'test': 'one', 'version': '0.1'})
        self.assertEqual(e.version, '0.1')

    def test_custom_version(self):
        e = event.Event({'test': 'one', 'version': '0.1'}, get_version=lambda x: '0.2')
        self.assertEqual(e.version, '0.2')

    def test_single_upgrade_mutation(self):
        @event.upgrade('0.1')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'
            event['version'] = '0.2'
        e = event.Event({'test': 'one', 'version': '0.1'})
        upgraded = set_name(e)
        self.assertEqual(upgraded.event['name'], 'test_single_upgrade_mutation')

    def test_single_upgrade_return(self):
        @event.upgrade('0.1')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'  # This should be overridden by the returned value
            return {'totally': 'different', 'dict': 'test'}
        e = event.Event({'test': 'one', 'version': '0.1'})
        upgraded = set_name(e)
        self.assertEqual(upgraded.event['totally'], 'different')

    def test_unmatched_version(self):
        @event.upgrade('0.2')
        def set_name(event):
            event['name'] = 'test_unmatched_version'
        e = event.Event({'test': 'one', 'version': '0.1'})
        upgraded = set_name(e)
        with self.assertRaises(KeyError):
            upgraded.event['name']

    def test_unmutated_event(self):
        @event.upgrade('0.1')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'
            event['version'] = '0.2'
        e = event.Event({'test': 'one', 'version': '0.1'})
        upgraded = set_name(e)
        with self.assertRaises(KeyError):
            e.event['name']

    def test_implicit_version_upgrade(self):
        @event.upgrade('0.1', '0.2')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'
        e = event.Event({'test': 'one', 'version': '0.1'})
        upgraded = set_name(e)
        self.assertEqual(upgraded.event['version'], '0.2')
        self.assertEqual(upgraded.version, '0.2')

    def test_version_not_upgraded(self):
        @event.upgrade('0.1')
        def set_name(event):
            event['name'] = 'test_single_upgrade_mutation'
        e = event.Event({'test': 'one', 'version': '0.1'})
        with self.assertRaises(KeyError):
            upgraded = set_name(e)

        
if __name__ == '__main__':
    unittest.main()


import copy
import inspect
from typing import Callable, Iterable, Any, List
from functools import wraps, reduce


VersionMethod = Callable[[dict], str]
VersionSetMethod = Callable[[dict, str], None]
FactoryMethod = Callable[[dict], Any]
UpgradeMethod = Callable[[dict], dict]


class Event:
    " Wrapper class for events "
    def __init__(self, event: dict, get_version: VersionMethod = None, set_version: VersionSetMethod = None):
        " get_version should be a function, not a method - there's no self provided in the call "
        self.event = event
        self.get_version: VersionMethod = get_version or (lambda x: str(x.get('version')))
        self.set_version: VersionSetMethod = set_version or (lambda x, y: x.update({'version': y}))

    @property
    def version(self) -> str:
        return getattr(self, 'get_version')(self.event)  # Phrased strangely because 'self.get_version()' will introduce a self

    @version.setter
    def version(self, value: str) -> None:
        return getattr(self, 'set_version')(self.event, value)  # Phrased strangely because 'self.set_version()' will introduce a self


def upgrade(version: str = '', new_version: str = ''):
    """
    Decorator to declare that a method is an upgrade method

    Takes version to be applied to, and optionally new version to set.
    """
    def upgrade_decorator(func: UpgradeMethod):
        @wraps(func)
        def wrapped_function(event: Event) -> Event:
            if event.version == version:
                newCopy = copy.deepcopy(event.event)
                ret = func(newCopy)
                newEvent = Event(ret and ret or newCopy, get_version=event.get_version, set_version=event.set_version)  # Allow return or mutate
                if new_version:
                    newEvent.version = new_version
                if newEvent.version == version:
                    raise KeyError('Upgrade did not change version number - {}'.format(version))
                return newEvent
            return event
        return wrapped_function
    return upgrade_decorator

class Upgrader:
    def __init__(self, upgrades: Iterable[Callable[[Event], Event]] = [], get_version: VersionMethod = None, set_version: VersionSetMethod = None, factory: FactoryMethod = None):
        " upgrades is a dict of version upgrade methods keyed on version "
        self.upgrades = dict({ (inspect.getclosurevars(u).nonlocals['version'], u) for u in upgrades })
        self.get_version = get_version
        self.set_version = set_version
        self.factory = factory

    def upgrade(self, event: dict, target_version: str = None) -> dict:
        " Apply all the upgrades to the event, to an optional target version "
        wrappedEvent = Event(event, get_version=self.get_version, set_version=self.set_version)
        while True:
            if target_version and (wrappedEvent.version == target_version):
                break
            u = self.upgrades.get(wrappedEvent.version, None)
            if u is None:
                break  # This should mean we've reached the end, may mean we have an aberrant version
            wrappedEvent = u(wrappedEvent)
        if (target_version is not None) and (target_version != wrappedEvent.version):
            raise KeyError("All upgrades done, target version {} doesn't match event version {}".format(target_version, wrappedEvent.version))
        return wrappedEvent.event

    def ingest(self, event: dict, target_version: str = None):  # Returns Any because factory can return anything
        " This is if you want to build an upgrader class then call it repeatedly for events "
        upgraded = self.upgrade(event, target_version=target_version)
        return (self.factory and self.factory(upgraded) or upgraded)


class Upgraders:  # Not yet done
    " Every method in here will be used as an upgrade method. Just be sure to decorate. "
    def upgraders(self) -> List[VersionMethod]:
        return [ method for name, method in inspect.getmembers(self, predicate=inspect.ismethod) if name != 'upgraders' ]



from setuptools import setup, find_packages  # type: ignore


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="pevl",
    version="0.0.1",
    packages=find_packages(),
    author="Kevin Littlejohn",
    author_email="kevin@littlejohn.id.au",
    description="Library to handle versions for events",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="event sourcing versioning",
    url="https://github.com/silarsis/pevl",
    python_requires=">=3.8"
)